#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   libs.py
@Time    :   2022/04/21 11:03:55
@Author  :   Thiago Nóbrega 
@Contact :   thiagonobrega@gmail.com
'''

import os
import time
import logging

import secrets

import solcx
from web3.exceptions import TransactionNotFound

def compile_all(contracts,output_values=["abi",'bin',"bin-runtime"],solc_version="0.8.13"):
    """
    """

    solcx.set_solc_version('0.8.13')
    compiled =  solcx.compile_files(contracts,
                                    output_values=output_values,
                                    solc_version=solc_version,
                                    optimize=True
                                    )

    return compiled


def compile_contract(contract_name,contracts,output_values=["abi",'bin',"bin-runtime","evm"],solc_version="0.8.13"):
# def compile_contract(contract_name,contracts,output_values=["abi",'bin',"evm"],solc_version="0.8.13"):
    """
    """
    compiled =  compile_all(contracts,
                            output_values=output_values,
                            solc_version=solc_version
                            )

    for k in compiled.keys():
        if k.endswith(str(os.sep + contract_name)):
            return compiled[k]
    
    raise Exception('This shouldnt happend')

def get_contract(contract_name,contracts_dicts):
    contract_name = contract_name.split('.')[0]
    for k in contracts_dicts.keys():
        if k.endswith(':'+contract_name):
            return contracts_dicts[k]
    
    raise Exception('This shouldnt happend')

def populate_file_list(dir,files):
    """
        Private method to populate a list with full path of the smartcontrats
    """
    lista = []
    for i in files:
        lista.append( os.path.join(dir,i) )
    return lista

def deploy_contract_besu(account,w3,contract_interface,chain_id,gas=500000):
   
   sc = w3.eth.contract( abi=contract_interface['abi'],
                           bytecode=contract_interface['bin']
                        )

   est_gas = sc.constructor().estimateGas()
   tx_const = sc.constructor().buildTransaction(get_tx_params(w3,est_gas,account,chain_id=chain_id))
   signed_tx = account.signTransaction(tx_const)
   tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)

   tx_receipt = None
   iter_count = 1

   tx_receipt = w3.eth.waitForTransactionReceipt(tx_hash)

   return tx_receipt['contractAddress']


def deploy_contract_besu_original(account,w3, contract_interface,gas=500000):
   
   sc = w3.eth.contract( abi=contract_interface['abi'],
                           bytecode=contract_interface['bin']
                        )

   #   'gas': 6612388,   
   # 'gasPrice': w3.eth.gasPrice
   tx_params = {'from': account.address,
               'nonce': w3.eth.getTransactionCount(account.address),
               'gas': gas,
               # 'gasPrice': 1100, #//ETH per unit of gas
               # BESU_MIN_GAS_PRICE=1337
               # 'gasLimit': '0x24A22' #//max number of gas units the tx is allowed to use
               }

   tx_const = sc.constructor().buildTransaction(tx_params)
   signed_tx = account.signTransaction(tx_const)
   tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)

   tx_receipt = None
   iter_count = 1

   # tx_receipt = w3.eth.waitForTransactionReceipt(tx_hash)

   while tx_receipt == None or iter_count < 30:
      logging.debug("Trying.. " + str(iter_count) + "/30 ...")
      time.sleep(2)
      try:
         tx_receipt = w3.eth.getTransactionReceipt(tx_hash)
         iter_count += 1
         return tx_receipt['contractAddress']
      except TransactionNotFound:
         iter_count += 1

   # tx_receipt['status'] 0x1 funcionou 0x0 nao foi feito o deploy
   if iter_count >= 29:
      logging.debug('Transacao lascou')
      
   return tx_receipt['contractAddress']

###
### gen privatekey
###

def gen_pk():
    """
        method to generate private keys to the block
    """
    priv = secrets.token_hex(32)
    return "0x" + priv



###
### Invoke the contract methods
###

def get_tx_params(w3,gas,account,chain_id=1337,min_gas_price='100'):
    # nonce = w3.eth.getTransactionCount(acount.address)
    nonce = w3.eth.getTransactionCount(w3.toChecksumAddress(account.address))
    tx_params = {'from': account.address,
                # 'to': contractAddress,
                'nonce': nonce,
                'gas': gas * 2,
                'gasPrice': w3.toWei(min_gas_price, 'gwei'), # defaul min gas price to besu (nao funciona w3.eth.gas_price)
                'chainId': chain_id
    }
    return tx_params


def invoke_contract(w3,account,chain_id,
                    smart_contract_method,*args):

    est_gas = smart_contract_method(*args).estimateGas()
    tx_params = get_tx_params(w3,est_gas,account,chain_id=chain_id)
    tx = smart_contract_method(args).buildTransaction(tx_params)
    signed_tx = account.signTransaction(tx)
    # send the transaction
    tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)
    receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    return receipt

def invoke_contract(w3,account,chain_id,
            smart_contract,method,*args):
    """
        invoke_contract(w3,account,61, st_service , 'set_db' ,(contract_addr) )
    """

    #get the gas needed
    est_gas = smart_contract.get_function_by_name(method)(*args).estimateGas()
    #gen the tc header
    tx_params = get_tx_params(w3,est_gas,account,chain_id=chain_id)
    #build the transaction
    tx = smart_contract.get_function_by_name(method)(*args).buildTransaction(tx_params)
    #sign the transatiopn
    signed_tx = account.signTransaction(tx)
    # send the transaction
    tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)
    receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    return receipt

def invoke_contract_experimental(w3,account,chain_id,
            smart_contract,method,*args):
    """
        invoke_contract(w3,account,61, st_service , 'set_db' ,(contract_addr) )
    """

    #get the gas needed
    est_gas = smart_contract.get_function_by_name(method)(*args).estimateGas()
    #gen the tc header
    tx_params = get_tx_params(w3,est_gas,account,chain_id=chain_id)
    #build the transaction
    tx = smart_contract.get_function_by_name(method)(*args).buildTransaction(tx_params)
    #sign the transatiopn
    signed_tx = account.signTransaction(tx)
    # send the transaction
    tx_hash = w3.eth.sendRawTransaction(signed_tx.rawTransaction)
    receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    return receipt , est_gas


from web3 import Web3
# from web3._utils.datatypes import Contract

class DarkPid:

    def __init__(self,pid_hash,ark_id,externa_pid_list,payload,owner) -> None:
        self.pid_hash = pid_hash
        self.ark = ark_id
        self.externa_pid_list = externa_pid_list
        self.payload = payload
        self.responsible = owner
    
    def to_dict(self):
        """
        Converts the attributes of the class object into a dictionary.

        Returns:
            dict: A dictionary representation of the class object's attributes.
        """
        return vars(self)
    
    def __is_bc_valid(bc_output):
        """
        Validates the structure and types of the provided output.

        Parameters:
            output (tuple): The output to be validated.

        Returns:
            bool: True if the output is valid, False otherwise.
        """

        if len(bc_output) != 10:
            return False

        if not isinstance(bc_output[0], bytes) or not isinstance(bc_output[1], str):
            return False

        if not isinstance(bc_output[2], list) or not isinstance(bc_output[3], list):
            return False

        if not isinstance(bc_output[4], int) or not isinstance(bc_output[5], list):
            return False

        if not isinstance(bc_output[6], int) or not isinstance(bc_output[7], tuple):
            return False

        if not isinstance(bc_output[8], str) or not isinstance(bc_output[9], str):
            return False

        return True


    def populateDark(dark_object,epid_db_contract):
        assert DarkPid.__is_bc_valid(dark_object) == True, "Invalid Blockchain Output"
        # TODO: VALIDADE epid_db_contract
        # assert type(epid_db_contract) == Contract, "epid_db_contract must be web3._utils.datatypes.Contract type"

        # populate external pids
        external_pids = []
        for ext_pid in dark_object[3]:
            ext_pid = Web3.toHex(ext_pid)
            # epid = epid_db.functions.get(ext_pid).call()
            get_func = epid_db_contract.get_function_by_signature('get(bytes32)')
            epid = get_func(ext_pid).call()

            pid_object = {'id': ext_pid, 
                            'schema:' : epid[3] , 'value' : epid[2], 
                            'owner:' : epid[-1]
                        }
            external_pids.append(pid_object)
        
        # deprecated
        # external_links = []
        # for ext_link in dark_object[5]:
        #     external_links.append(ext_link)

        pid_hash_id = Web3.toHex(dark_object[0])
        pid_ark_id = dark_object[1]
        payload = dark_object[-2]
        owner = dark_object[-1]
        

        return DarkPid(pid_hash_id,pid_ark_id,external_pids,payload,owner)

        


class ExeternalPid:
    def __init__(self,id_hash,schema,value,creator) -> None:
        self.id = id_hash
        self.schema = schema
        self.pid_str_value = value
        self.creator_addr = creator

    def to_dict(self):
        """
        Converts the attributes of the class object into a dictionary.

        Returns:
            dict: A dictionary representation of the class object's attributes.
        """
        return vars(self)


#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   setup.py
@Time    :   2022/05/13 10:20:06
@Author  :   Thiago Nóbrega 
@Contact :   thiagonobrega@gmail.com
'''

import os
import logging
import ast
import configparser

from web3 import Web3, IPCProvider
from web3.middleware import geth_poa_middleware

# from eth_tester import PyEVMBackend
from web3.providers.eth_tester import EthereumTesterProvider


class DarkGateway:

    def __init__(self, blockchain_net_name: str,
                 blockchain_config: configparser.SectionProxy):
        
        #TODO: MODIFY CONSTRUCTOR PARAMATERS
        assert type(blockchain_net_name) == str, "blockchain_net_name must be str type"
        assert type(blockchain_config) == configparser.SectionProxy, "blockchain_config must be configparser.SectionProxy type"
        

        ### w3dark config parameter
        self.__blockchain_net_name = blockchain_net_name
        self.__blockchain_config = blockchain_config # could be removed

        ### blockchain exec params
        self.__chain_id = int(blockchain_config['chain_id'])
        self.__min_gas_price = int(blockchain_config['min_gas_price'])
        self.__pk = blockchain_config['account_priv_key'] #FIXME: Possible security risk
        self.min_gas_price = '100' #FIXME: add configuration parameter

        ### important variables
        self.w3 =  self.__class__.load_blockchain_driver(blockchain_net_name,blockchain_config)
        self.deployed_contracts_dict = None

        ### account
        self.__account = self.w3.eth.account.privateKeyToAccount(self.__pk)
        # endereco da autoridade
        #TODO: modelar utilizar  multiplas autoridades
        #TODO: colocar como variavel de configuração
        #FIXME: quando for necessario utilizar multiplas autoridades
        self.authority_addr = "0xf17f52151EbEF6C7334FAD080c5704D77216b732"
  
    def load_deployed_smart_contracts(self,deployed_contracts_config:configparser.ConfigParser):
        """
            Load the deployed smart contracts
            - Ity is essential notice that it is important to configure the smart contract
        """
        assert type(deployed_contracts_config) == configparser.ConfigParser, "deployed_contracts_config must be configparser.ConfigParser type"
        # self.__deployed_contracts_config = deployed_contracts_config

        contracts_dict = {}
        for k in list(deployed_contracts_config.keys()):
            if k != 'DEFAULT':
                addr = deployed_contracts_config[k]['addr']
                c_abi = ast.literal_eval(deployed_contracts_config[k]['abi'])['abi']
                contracts_dict[k] = self.w3.eth.contract(address=addr, abi=c_abi)

        # TODO: CHECK IF CONTRANCT DICT ARE EMPTY
        self.deployed_contracts_dict = contracts_dict
        
    def is_deployed_contract_loaded(self):
        return self.deployed_contracts_dict != None
    
    def get_exec_parameters(self):
        """
            Return the blockchain execution parameters
            - chain_id
            - ming_gas_price
            - pk
        """
        return self.__chain_id,self.__min_gas_price,self.__pk
    
    ####
    #### sing and send parameters
    ####

    def get_tx_params(self,gas):
        """
            Method employed to retrive the BC tx params
        """
                    #   w3,gas,account,chain_id=1337,min_gas_price='100'):
        # nonce = w3.eth.getTransactionCount(acount.address)
        nonce = self.w3.eth.getTransactionCount(self.w3.toChecksumAddress(self.__account.address))
        tx_params = {'from': self.__account.address,
                    # 'to': contractAddress,
                    'nonce': nonce,
                    'gas': gas * 2,
                    'gasPrice': self.w3.toWei(self.min_gas_price, 'gwei'), # defaul min gas price to besu (nao funciona w3.eth.gas_price)
                    'chainId': self.__chain_id
        }
        return tx_params
    
    def signTransaction(self,smart_contract,method,*args):
        #get the gas needed
        est_gas = smart_contract.get_function_by_name(method)(*args).estimateGas()
        tx_params = self.get_tx_params(est_gas)
        #build the transaction
        tx = smart_contract.get_function_by_name(method)(*args).buildTransaction(tx_params)
        signed_tx = self.__account.signTransaction(tx)
        return signed_tx
    



    ###
    ### private methods
    ###
    @staticmethod
    def load_blockchain_driver(blockchain_net_name: str,blockchain_config: configparser.SectionProxy) -> Web3:
        """
            Load the blockchain driver.

            The drive is used to connect the application to the blockchain.
            The configuration is defined in config.ini file.
        """
        assert type(blockchain_net_name) == str, "blockchain_net_name must be str type"
        assert type(blockchain_config) == configparser.SectionProxy, "blockchain_config must be configparser.SectionProxy type"
        
        #debug
        # logging.info(config_file)
        w3 = None

        if blockchain_net_name == 'EthereumTesterPyEvm':
            raise(Exception("Not Suported"))
            # return Web3(EthereumTesterProvider(PyEVMBackend()))
        elif 'dpi-' in blockchain_net_name:
            # blockchain_config = config[blockchain_net_name]
            # blockchain_config['url']
            w3 = Web3(Web3.HTTPProvider(blockchain_config['url']))
            w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            # return w3
        else:
            raise RuntimeError('This shouldnt happend :: Not implemented')
        
        return w3




import os
import time
import logging

import secrets

import solcx
from web3.exceptions import TransactionNotFound


from ..gateway import DarkGateway


def invoke_contract_sync(dark_gateway: DarkGateway,
                    signed_tx):
        
        assert type(dark_gateway) == DarkGateway, "dark_gateway must be a DarkGateway object"

        # signed_tx = dark_gateway.signTransaction(smart_contract,method,*args)
        tx_hash = dark_gateway.w3.eth.sendRawTransaction(signed_tx.rawTransaction)
        receipt = dark_gateway.w3.eth.wait_for_transaction_receipt(tx_hash)
        return receipt, tx_hash


#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   setup.py
@Time    :   2022/05/13 10:20:06
@Author  :   Thiago Nóbrega 
@Contact :   thiagonobrega@gmail.com
'''
import os
import configparser
import logging
import ast

from web3 import Web3, IPCProvider
from web3.middleware import geth_poa_middleware

# from eth_tester import PyEVMBackend
from web3.providers.eth_tester import EthereumTesterProvider

from .libs import compile_all,get_contract,deploy_contract_besu,populate_file_list,get_tx_params,invoke_contract

#TODO: Definir melhor
PROJECT_ROOT = os.path.dirname(os.path.dirname(__file__))

config_file = os.path.join(PROJECT_ROOT,'config.ini')
deployed_contracts_config = os.path.join(PROJECT_ROOT,'deployed_contracts.ini')


def load_blockchain_driver(blockchain_net_name: str,blockchain_config: configparser.SectionProxy):
    """
        Load the blockchain driver.

        The drive is used to connect the application to the blockchain.
        The configuration is defined in config.ini file.
    """
    assert type(blockchain_net_name) == str, "blockchain_net_name must be str type"
    assert type(blockchain_config) == configparser.SectionProxy, "blockchain_config must be configparser.SectionProxy type"
    
    #debug
    logging.info(config_file)
    # print(config_file)

    # config = configparser.ConfigParser()
    # config.read(config_file)
    # blockchain_net = config['base']['blockchain_net']

    if blockchain_net_name == 'EthereumTesterPyEvm':
        raise(Exception("Not Suported"))
        # return Web3(EthereumTesterProvider(PyEVMBackend()))
    elif 'dpi-' in blockchain_net_name:
        # blockchain_config = config[blockchain_net_name]
        # blockchain_config['url']
        w3 = Web3(Web3.HTTPProvider(blockchain_config['url']))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        return w3
    else:
        raise RuntimeError('This shouldnt happend :: Not implemented')

######
###### compile
######

def compile():
    logging.info("> Compiling Contracts")
    config = configparser.ConfigParser()
    config.read(config_file)

    DAPP_ROOT = os.path.join(PROJECT_ROOT, 'Dapp')
    LIB_PATH = os.path.join(DAPP_ROOT, config['smartcontracts']['lib_dir'])
    UTIL_PATH = os.path.join(DAPP_ROOT, config['smartcontracts']['util_dir'])
    DB_PATH = os.path.join(DAPP_ROOT, config['smartcontracts']['db_dir'])
    SERVICE_PATH = os.path.join(DAPP_ROOT, config['smartcontracts']['service_dir'])

    libs_path = populate_file_list(LIB_PATH,config['smartcontracts']['lib_files'].split())
    utils_path = populate_file_list(UTIL_PATH,config['smartcontracts']['utils_files'].split())
    dbs_paths = populate_file_list(DB_PATH,config['smartcontracts']['dbs_files'].split())
    services_path = populate_file_list(SERVICE_PATH,config['smartcontracts']['service_files'].split())
    
    compiled_contracts = compile_all(services_path + dbs_paths + utils_path + libs_path)
    logging.info("")
    return compiled_contracts

######
###### deploy
######

def deploy(w3,account,contract_name,compiled_contracts,chain_id):
    contract_interface = get_contract(contract_name,compiled_contracts)
    addr = deploy_contract_besu(account,w3,contract_interface,chain_id)
    # contracts_dicts[contract_name] = 
    return [addr,contract_interface]

def deploy_contracts(w3,compiled_contracts):
    config = configparser.ConfigParser()
    config.read(config_file)

    blockchain_net = config['base']['blockchain_net']
    logging.info("> Deployin Contracts to " + blockchain_net)

    if blockchain_net == 'EthereumTesterPyEvm':
        #transferindo fundo para uma conta nova
        pk = config['EthereumTesterPyEvm']['account_priv_key']
        chain_id = int(config['EthereumTesterPyEvm']['chain_id'])
        min_gas_price = int(config['EthereumTesterPyEvm']['min_gas_price'])
        # account = w3.eth.account.create(pk)
        account = w3.eth.account.privateKeyToAccount(pk)


        tx = { "from": w3.eth.accounts[0], 
                'to': account.address, 
                "value": Web3.toWei(10000, "ether"), 
                'gas': 21000 
        }
        w3.eth.send_transaction(tx)
        
    elif 'dpi-' in blockchain_net:
        pk = config[blockchain_net]['account_priv_key']
        account = w3.eth.account.privateKeyToAccount(pk)
        chain_id = int(config[blockchain_net]['chain_id'])
        #TODO: passar como parametro para os metodos de deploy
        min_gas_price = int(config[blockchain_net]['min_gas_price'])
    else:
        raise RuntimeError('This shouldnt happend :: Not implemented')
    
    acc_balance = str(Web3.fromWei(w3.eth.get_balance(account.address),'ether'))
    logging.info("    using account : " + account.address )
    logging.info("    account initial balance : " + acc_balance )

    # lista = config['smartcontracts']['lib_files'].split() +\
    lista = config['smartcontracts']['utils_files'].split() +\
            config['smartcontracts']['dbs_files'].split() +\
            config['smartcontracts']['service_files'].split()

    deployed_contract_dict = {}
    
    for contract_name in lista:
        if contract_name not in ['Entities.sol' , 'NoidProvider.sol']:
            logging.info("    deploying : " + str(contract_name) + "..." )
            deployed_contract_dict[str(contract_name)] = deploy(
                                                        w3,account,
                                                        contract_name,
                                                        compiled_contracts,
                                                        chain_id)
            
    logging.info("    deployed : " + str(len(lista)) + " contracts" )
    acc_balance = str(Web3.fromWei(w3.eth.get_balance(account.address),'ether'))
    logging.info("    account initial balance : " + acc_balance )
    logging.info("")
    return deployed_contract_dict

######
###### configuration
######

def configure_env(w3,deployed_contract_dict):
    config = configparser.ConfigParser()
    config.read(config_file)

    configured_contracts = {}
    
    if config['base']['blockchain_net'] == 'EthereumTesterPyEvm':
        blockchain_conf = config['EthereumTesterPyEvm']
    elif 'dpi-' in config['base']['blockchain_net']:
        blockchain_conf = config[config['base']['blockchain_net']]
    else:
        raise RuntimeError('This shouldnt happend :: Not implemented')
    
    chain_id = int(blockchain_conf['chain_id'])
    min_gas_price = int(blockchain_conf['min_gas_price'])
    pk = blockchain_conf['account_priv_key']

    ############## setup
    logging.info("> Configure env...")
    account = w3.eth.account.privateKeyToAccount(pk)
    acc_balance = str(Web3.fromWei(w3.eth.get_balance(account.address),'ether'))
    logging.info("    using account : " + account.address )
    logging.info("    account initial balance : " + acc_balance )

    ### Authorities Service
    logging.info("    Configuring AuthoritiesService:")
    auth_db_addr = deployed_contract_dict['AuthoritiesDB.sol'][0]
    contract_addr = deployed_contract_dict['AuthoritiesService.sol'][0]
    contract_interface = deployed_contract_dict['AuthoritiesService.sol'][1]
    auth_service = w3.eth.contract(address=contract_addr, abi=contract_interface["abi"])
    invoke_contract(w3,account,chain_id, auth_service , 'set_db' ,(auth_db_addr) )
    logging.info("        - db configured")
    logging.info("        - AuthoritiesService configured")
    configured_contracts['AuthoritiesService'] = auth_service
    

    ### pid db
    logging.info("    Configuring D-pi PiD Database:")
    # uuid_provider_addr = deployed_contract_dict['UUIDProvider.sol'][0]
    contract_addr = deployed_contract_dict['PidDB.sol'][0]
    contract_interface = deployed_contract_dict['PidDB.sol'][1]
    pid_db = w3.eth.contract(address=contract_addr, abi=contract_interface["abi"])
    # invoke_contract(w3,account,chain_id, pid_db , 'set_uuid_provider' ,(uuid_provider_addr) )
    # logging.info("        - uuid provider configured")
    configured_contracts['PidDB'] = pid_db
    
    ### Search TermService
    logging.info("    Configuring SearchTermService:")
    st_db_addr = deployed_contract_dict['SearchTermDB.sol'][0]
    contract_addr = deployed_contract_dict['SearchTermService.sol'][0]
    contract_interface = deployed_contract_dict['SearchTermService.sol'][1]
    st_service = w3.eth.contract(address=contract_addr, abi=contract_interface["abi"])
    invoke_contract(w3,account,chain_id, st_service , 'set_db' ,(st_db_addr) )
    logging.info("        - db configured")
    configured_contracts['SearchTermService'] = st_service

    #### ExternalPID Service
    logging.info("    Configuring ExternalPIDService:")
    epid_db_addr = deployed_contract_dict['ExternalPidDB.sol'][0]
    contract_addr = deployed_contract_dict['ExternalPIDService.sol'][0]
    contract_interface = deployed_contract_dict['ExternalPIDService.sol'][1]
    epid_service = w3.eth.contract(address=contract_addr, abi=contract_interface["abi"])
    invoke_contract(w3,account,chain_id, epid_service , 'set_db' ,(epid_db_addr) )
    logging.info("        - db configured")
    configured_contracts['ExternalPIDService'] = epid_service
    

    # D-pi PID Service
    logging.info("    Configuring D-pi PIDService:")
    pid_db_addr = deployed_contract_dict['PidDB.sol'][0]
    contract_addr = deployed_contract_dict['PIDService.sol'][0]
    contract_interface = deployed_contract_dict['PIDService.sol'][1]
    pid_service = w3.eth.contract(address=contract_addr, abi=contract_interface["abi"])
    invoke_contract(w3,account,chain_id, pid_service , 'set_db' ,(pid_db_addr) )
    logging.info("        - db configured")
    invoke_contract(w3,account,chain_id, pid_service , 'set_externalpid_service' ,(epid_service.address) )
    logging.info("        - ExternalPIDService configured")
    invoke_contract(w3,account,chain_id, pid_service , 'set_searchterm_service' ,(st_service.address) )
    logging.info("        - SearchTermService configured")
    invoke_contract(w3,account,chain_id, pid_service , 'set_auth_service' ,(auth_service.address) )
    logging.info("        - authoritiesService configured")
    configured_contracts['PIDService'] = pid_service

    ### all set
    acc_balance = str(Web3.fromWei(w3.eth.get_balance(account.address),'ether'))
    logging.info("    account final balance : " + acc_balance )
    logging.info("All set... Services configured.")
    logging.info("")
    return configured_contracts

###
###
### 
def get_exec_parameters():
    
    config = configparser.ConfigParser()
    config.read(config_file)

    bc_net = config['base']['blockchain_net'] 
    blockchain_conf = config[bc_net]

    chain_id = int(blockchain_conf['chain_id'])
    min_gas_price = int(blockchain_conf['min_gas_price'])
    pk = blockchain_conf['account_priv_key']
    return chain_id,min_gas_price,pk

def save_smart_contract(deployed_contracts):
    """
        Save the deployed smart contracts
        - inputs the deploed_contracts dict

        - It is essential to configure the contract prior to its usage
        - Please only use this method to save configured contracts
    """
    config = configparser.ConfigParser()

    for k in deployed_contracts.keys():
        addr = deployed_contracts[k][0] # addr
        v2 = deployed_contracts[k][1].copy()
        del v2['bin']
        del v2['bin-runtime']
        config[k] = { 'addr': addr , 'abi' : v2}

    with open(deployed_contracts_config, 'w') as configfile:
        config.write(configfile)

def load_deployed_smart_contracts(w3):
    """
        Save the deployed smart contracts
        - Ity is essential notice that it is important to configure the smart contract
    """
    

    c2 = configparser.ConfigParser()
    c2.read(deployed_contracts_config)

    contracts_dict = {}
    for k in list(c2.keys()):
        if k != 'DEFAULT':
            addr = c2[k]['addr']
            c_abi = ast.literal_eval(c2[k]['abi'])['abi']
            contracts_dict[k] = w3.eth.contract(address=addr, abi=c_abi)
    
    return contracts_dict



def setup_from_scratch():
    w3 = load_blockchain_driver()
    compiled_contracts = compile()
    deployed_contract_dict = deploy_contracts(w3,compiled_contracts)
    configured_contracts = configure_env(w3,deployed_contract_dict)



if __name__ == '__main__':

    import logging
    logging.basicConfig(level=logging.INFO)

    w3 = load_blockchain_driver()
    compiled_contracts = compile()
    deployed_contract_dict = deploy_contracts(w3,compiled_contracts)
    configured_contracts = configure_env(w3,deployed_contract_dict)

    #
    chain_id,min_gas_price,pk = get_exec_parameters()
    account = w3.eth.account.privateKeyToAccount(pk)


    pidService = configured_contracts['PIDService']
    r = invoke_contract(w3,account,chain_id, pidService , 'assingUUID' )
    pidDB = configured_contracts['PidDB']

    ###
    pid_db = compiled_contracts['/mnt/d/Dados/OneDrive/IBICT/workspace/D-pi/Dapp/db/PidDB.sol:PidDB']
    addr = configured_contracts['PidDB'].address
    
    pidDB = w3.eth.contract(address=addr, abi=pid_db["abi"])
    ###
    
    pidDB.get_function_by_name('count').call()
    
    account.address
    pidDB.caller.count()
    Web3.toHex(pidDB.caller.get_by_index(0))
    pid_id = r['logs'][0]['topics'][1][:16]

    pid_object = pidDB.caller.get(pid_id)

    #olhar no Entities libs as posicoes 0 ~ 9
    print('uuid='+Web3.toHex(pid_object[0]))
    print("owner="+pid_object[8])
    


import setuptools
import codecs
import os


local = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(local, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()


# This gets deployed when a new release is made by github actions
VERSION = '0.0.2a3'
# VERSION = '0.0.1'

# CHANGEME VARS
PACKAGE_NAME = "dark-gateway"
DESCRIPTION = 'dARK Gateway libs'
LONG_DESCRIPTION = 'dARK web3 libs to connect applications to the dARK'
AUTHOR_NAME = "Thiago Nóbrega"
AUTHOR_EMAIL = "thiagonobrega@gmail.com"
PROJECT_URL = "https://github.com/dark-pid/dark-web3-lib"
PROJECT_KEYWORDS = ['ARK', 'dARK', 'PID', 'web3']
# required 3rd party tools used by your package
# REQUIRED_PACKAGES = ['pandas'] # unsuser

# Read more about classifiers at https://pypi.org/classifiers/
CLASSIFIERS = [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: OS Independent"]

setuptools.setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author=AUTHOR_NAME,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    url = PROJECT_URL,
    long_description=long_description,
    long_description_content_type="text/markdown",
    # install_requires=REQUIRED_PACKAGES,
    packages=setuptools.find_packages(),
    keywords=PROJECT_KEYWORDS,
    classifiers=CLASSIFIERS,
)


from .gateway import DarkGateway
from .util import invoke_contract_sync
from .pid_modules import DarkPid


class DarkMap:
    
    def __init__(self, dark_gateway: DarkGateway):
        assert type(dark_gateway) == DarkGateway, "dark_gateway must be a DarkGateway object"
        assert dark_gateway.is_deployed_contract_loaded() == True, "dark_gateway must be loaded with deployed contracts"

        #dark gatewar
        self.gw = dark_gateway

        ##
        ## dARK SmartContracts
        ##

        # databases for query
        self.dpid_db = dark_gateway.deployed_contracts_dict['PidDB.sol']
        self.epid_db = dark_gateway.deployed_contracts_dict['ExternalPidDB.sol']
        self.sete_db = dark_gateway.deployed_contracts_dict['SearchTermDB.sol']
        # authorities db to configuration
        self.auth_db = dark_gateway.deployed_contracts_dict['AuthoritiesDB.sol']
        #dARK services
        self.dpid_service = dark_gateway.deployed_contracts_dict['PIDService.sol']
        self.epid_service = dark_gateway.deployed_contracts_dict['ExternalPIDService.sol']
        self.sets_service = dark_gateway.deployed_contracts_dict['SearchTermService.sol']
        self.auth_service = dark_gateway.deployed_contracts_dict['AuthoritiesService.sol']
    
    
    ###
    ### Request PID
    ###

    def request_pid_hash(self):
        """
            Request a PID and return the hash (address) of the PID
        """
        signed_tx = self.gw.signTransaction(self.dpid_service , 'assingID', self.gw.authority_addr)
        r_tx, receipt = invoke_contract_sync(self.gw,signed_tx)
        dark_id = r_tx['logs'][0]['topics'][1]
        return dark_id
    
    def request_pid(self):
        """
            Request a PID and return the ark of the PID
        """
        return self.convert_pid_hash_to_ark(self.request_pid_hash())
    
    def convert_pid_hash_to_ark(self,dark_pid_hash):
        """
            Convert the dark_pid_hash to a ARK identifier
        """
        return self.dpid_db.caller.get(dark_pid_hash)[1]
    
    ###
    ### Onchain core queries
    ###

    def get_pid_by_hash(self,dark_id):
        """
            Retrieves a persistent identifier (PID) by its hash value.

            Parameters:
                dark_id (str): The hash value of the PID.

            Returns:
                str: The PID associated with the given hash value.

            Raises:
                AssertionError: If the dark_id does not start with '0x'.
        """
        assert dark_id.startswith('0x'), "id is not hash"
        dark_object = self.dpid_db.caller.get(dark_id)
        return DarkPid.populateDark(dark_object,self.epid_db)

    def get_pid_by_ark(self,dark_id):
        """
            Retrieves a persistent identifier (PID) by its ARK (Archival Resource Key) identifier.

            Parameters:
                dark_id (str): The ARK identifier of the PID.

            Returns:
                str: The PID associated with the given ARK identifier.
        """
        dark_object = self.dpid_db.caller.get_by_noid(dark_id)
        return DarkPid.populateDark(dark_object,self.epid_db)


